"""Integration tests for Gandi"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.gandi import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class GandiRPCProviderTests(TestCase, IntegrationTests):
    """TestCase for Gandi on RPC"""
    Provider = Provider
    provider_name = 'gandi'
    domain = 'reachlike.ca'
    provider_variant = 'RPC'

    def _test_parameters_overrides(self):
        return {'api_protocol': 'rpc'}


class GandiRESTProviderTests(TestCase, IntegrationTests):
    """TestCase for Gandi on REST API"""
    Provider = Provider
    provider_name = 'gandi'
    domain = 'reachfactory.ca'
    provider_variant = 'REST'

    def _filter_headers(self):
        return ['X-Api-Key']

    def _test_parameters_overrides(self):
        return {'api_protocol': 'rest'}



"""Module provider for OnApp"""
from __future__ import absolute_import
import json
import logging

import requests
from requests.auth import HTTPBasicAuth

from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = []


def provider_parser(subparser):
    """Configure provider parser for OnApp"""
    subparser.description = '''
        The OnApp provider requires your OnApp account\'s email address and
        API token, which can be found on your /profile page on the Control Panel interface.
        The server is your dashboard URL, with format like https://dashboard.youronapphost.org'''
    subparser.add_argument(
        '--auth-username', help='specify email address of the OnApp account')
    subparser.add_argument(
        '--auth-token', help='specify API Key for the OnApp account')
    subparser.add_argument(
        '--auth-server', help='specify URL to the OnApp Control Panel Server')


class Provider(BaseProvider):
    """Provider class for OnApp"""
    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.domain_id = None

        if not self._get_provider_option('auth_username'):
            raise Exception('Error, OnApp Email Address is not defined')
        if not self._get_provider_option('auth_token'):
            raise Exception('Error, OnApp API Key is not defined')
        if not self._get_provider_option('auth_server'):
            raise Exception('Error, OnApp Control Panel URL is not defined')

        self.session = requests.Session()

    def _authenticate(self):
        domain = self.domain

        zones = self._get('/dns_zones.json')
        for zone in zones:
            if zone['dns_zone']['name'] == domain:
                self.domain_id = zone['dns_zone']['id']
                break

        if self.domain_id is None:
            raise Exception(
                'Could not find {0} in OnApp DNS Zones'.format(domain))

    def _create_record(self, rtype, name, content):
        data = {
            'name': self._relative_name(name),
            'type': rtype,
            self._key_for_record_type(rtype): content
        }

        ttl = self._get_lexicon_option('ttl')
        if ttl:
            data['ttl'] = "{0}".format(ttl)

        result = self._post(
            '/dns_zones/{0}/records.json'.format(self.domain_id), {'dns_record': data})
        LOGGER.debug('create_record: %s', result)

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        records = []

        response = self._get(
            '/dns_zones/{0}/records.json'.format(self.domain_id))
        for record_type in response['dns_zone']['records']:

            # For now we do not support other RR types so we ignore them, also see
            # _key_for_record_type
            if record_type not in ('A', 'AAAA', 'CNAME', 'TXT'):
                continue

            if rtype and record_type != rtype:
                continue

            for record in response['dns_zone']['records'][record_type]:
                record = record['dns_record']

                if name and record['name'] != self._relative_name(name):
                    continue

                record_content = record[self._key_for_record_type(record_type)]
                if content and record_content != content:
                    continue

                records.append({
                    'id': record['id'],
                    'name': self._full_name(record['name']),
                    'type': record['type'],
                    'ttl': record['ttl'],
                    'content': record_content
                })

        LOGGER.debug('list_records: %s', records)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            existing = self._guess_record(rtype, name)
            identifier = existing['id']

        ttl = self._get_lexicon_option('ttl')

        if not name or not ttl:
            if not existing:
                existing = self._get(
                    '/dns_zones/{0}/records/{1}.json'.format(self.domain_id, identifier))
            if not name:
                name = existing['name']
            if not ttl:
                ttl = existing['ttl']

        request = {
            'name': self._relative_name(name),
            'ttl': '{0}'.format(ttl),
            self._key_for_record_type(rtype): content
        }

        result = self._put('/dns_zones/{0}/records/{1}.json'.format(
            self.domain_id, identifier), {'dns_record': request})
        LOGGER.debug('update_record: %s', result)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        deletion_ids = []

        if not identifier:
            records = self._list_records(rtype, name, content)
            deletion_ids = [record['id'] for record in records]
        else:
            deletion_ids.append(identifier)

        for one_id in deletion_ids:
            self._delete(
                '/dns_zones/{0}/records/{1}.json'.format(self.domain_id, one_id))

        LOGGER.debug('delete_record: %s', True)

        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        target = self._get_provider_option('auth_server') + url

        body = ''
        if data is not None:
            body = json.dumps(data)

        auth = HTTPBasicAuth(self._get_provider_option(
            'auth_username'), self._get_provider_option('auth_token'))

        request = requests.Request(
            action, target, data=body, headers=headers, params=query_params, auth=auth)
        prepared_request = self.session.prepare_request(request)

        result = self.session.send(prepared_request)
        result.raise_for_status()

        if result.text:
            return result.json()
        return None

    def _key_for_record_type(self, record_type):  # pylint: disable=no-self-use
        if record_type in ('A', 'AAAA'):
            return 'ip'
        if record_type == 'CNAME':
            return 'hostname'
        if record_type == 'TXT':
            return 'txt'
        if record_type in ('MX', 'NS', 'SOA', 'SRV', 'LOC'):
            raise Exception(
                '{0} record type is not supported in the OnApp Provider'.format(record_type))
        raise Exception(
            '{0} record type is unknown'.format(record_type))

    def _guess_record(self, rtype, name=None, content=None):
        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) == 1:
            return records[0]
        if len(records) > 1:
            raise Exception(
                'Identifier was not provided and several existing records '
                'match the request for {0}/{1}'.format(rtype, name))
        else:
            raise Exception(
                'Identifier was not provided and no existing records '
                'match the request for {0}/{1}'.format(rtype, name))



"""Module provider for DNSMadeEasy"""
from __future__ import absolute_import
import hmac
import json
import logging
from builtins import bytes
from email.utils import formatdate
from hashlib import sha1

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['dnsmadeeasy']


def provider_parser(subparser):
    """Configure provider parser for DNSMadeEasy"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication")
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for DNSMadeEasy"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = self._get_provider_option(
            'api_endpoint') or 'https://api.dnsmadeeasy.com/V2.0'

    def _authenticate(self):

        try:
            payload = self._get('/dns/managed/name',
                                {'domainname': self.domain})
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                payload = {}
            else:
                raise

        if not payload or not payload['id']:
            raise Exception('No domain found')

        self.domain_id = payload['id']

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        record = {
            'type': rtype,
            'name': self._relative_name(name),
            'value': content,
            'ttl': self._get_lexicon_option('ttl')
        }
        payload = {}
        try:
            payload = self._post(
                '/dns/managed/{0}/records/'.format(self.domain_id), record)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code != 400:
                raise

                # http 400 is ok here, because the record probably already exists
        LOGGER.debug('create_record: %s', 'name' in payload)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_query = {}
        if rtype:
            filter_query['type'] = rtype
        if name:
            filter_query['recordName'] = self._relative_name(name)
        payload = self._get(
            '/dns/managed/{0}/records'.format(self.domain_id), filter_query)

        records = []
        for record in payload['data']:
            processed_record = {
                'type': record['type'],
                'name': '{0}.{1}'.format(record['name'], self.domain),
                'ttl': record['ttl'],
                'content': record['value'],
                'id': record['id']
            }

            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if content:
            records = [
                record for record in records if record['content'].lower() == content.lower()]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {
            'id': identifier,
            'ttl': self._get_lexicon_option('ttl')
        }

        if name:
            data['name'] = self._relative_name(name)
        if content:
            data['value'] = content
        if rtype:
            data['type'] = rtype

        self._put(
            '/dns/managed/{0}/records/{1}'.format(self.domain_id, identifier), data)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self._delete(
                '/dns/managed/{0}/records/{1}'.format(self.domain_id, record_id))

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'x-dnsme-apiKey': self._get_provider_option('auth_username')
        }
        default_auth = None

        # Date string in HTTP format e.g. Sat, 12 Feb 2011 20:59:04 GMT
        request_date = formatdate(usegmt=True)

        hashed = hmac.new(bytes(self._get_provider_option('auth_token'), 'ascii'),
                          bytes(request_date, 'ascii'), sha1)

        default_headers['x-dnsme-requestDate'] = request_date
        default_headers['x-dnsme-hmac'] = hashed.hexdigest()

        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers,
                                    auth=default_auth)
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        # PUT and DELETE actions dont return valid json.
        if action in ['DELETE', 'PUT']:
            return response.text
        return response.json()



""""Test for rackspace implementation of the lexicon interface"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.rackspace import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class RackspaceProviderTests(TestCase, IntegrationTests):
    """Tests the rackspace provider"""

    Provider = Provider
    provider_name = 'rackspace'
    domain = 'capsulecd.com'

    def _filter_post_data_parameters(self):
        return ['auth']

    def _filter_headers(self):
        return ['X-Auth-Token']

    # Rackspace does not provide a sandbox API; actual credentials are required
    # Replace the auth_account, auth_username and auth_api_key as well as the
    # domain above with an actual domain you have added to Rackspace to
    # regenerate the fixtures
    def _test_parameters_overrides(self):
        # Set this to 1 if you are making new recordings and set to 0 when
        # finished to make sure we don't actually sleep and waste time
        # when we are replaying. Rackspace API calls are async, so you place
        # the initial request and then make update calls to see if the action
        # is complete and has a response.
        return {
            'sleep_time': '0',
        }

    def _test_fallback_fn(self):
        return lambda x: 'placeholder_' + x if x != 'auth_token' else None



"""Base class for provider integration tests"""
# pylint: disable=missing-docstring
import os
from builtins import object
from functools import wraps
from importlib import import_module

import pytest
import vcr
from lexicon.config import ConfigResolver, ConfigSource, DictConfigSource
from lexicon.providers.base import Provider as BaseProvider


# Configure VCR. Parameter record_mode depends on the LEXICON_LIVE_TESTS environment variable value.
RECORD_MODE = 'none'
if os.environ.get('LEXICON_LIVE_TESTS', 'false') == 'true':
    RECORD_MODE = 'once'
PROVIDER_VCR = vcr.VCR(
    cassette_library_dir=os.environ.get('LEXICON_VCRPY_CASSETTES_PATH',
                                        'tests/fixtures/cassettes'),
    record_mode=RECORD_MODE,
    decode_compressed_response=True
)

# Prepare custom decorator: it will start a casette in relevant folder for current provider,
# and using the name of the test method as the cassette's name.
def _vcr_integration_test(decorated):
    @wraps(decorated)
    def wrapper(self):
        # pylint: disable=protected-access
        with PROVIDER_VCR.use_cassette(
                self._cassette_path('IntegrationTests/{0}.yaml'
                                    .format(decorated.__name__)),
                filter_headers=self._filter_headers(),
                before_record_response=self._filter_response,
                filter_query_parameters=self._filter_query_parameters(),
                filter_post_data_parameters=self._filter_post_data_parameters()):
            decorated(self)
        # pylint: enable=protected-access
    return wrapper


class EngineOverrideConfigSource(ConfigSource):  # pylint: disable=too-few-public-methods
    """Config source to override some provider parameters during tests"""
    def __init__(self, overrides):
        super(EngineOverrideConfigSource, self).__init__()
        self.overrides = overrides

    def resolve(self, config_key):
        # We extract the key from existing namespace.
        config_key = config_key.split(':')[-1]
        return self.overrides.get(config_key)


class FallbackConfigSource(ConfigSource):  # pylint: disable=too-few-public-methods
    """Config source to provider fallback to provider parameters during tests"""
    def __init__(self, fallback_fn):
        super(FallbackConfigSource, self).__init__()
        self.fallback_fn = fallback_fn

    def resolve(self, config_key):
        config_key = config_key.split(':')
        if not config_key[-2] == 'lexicon':
            return self.fallback_fn(config_key[-1])

        return None


class IntegrationTests(object):  # pylint: disable=useless-object-inheritance,too-many-public-methods
    """
    https://stackoverflow.com/questions/26266481/pytest-reusable-tests-for-different-implementations-of-the-same-interface  # pylint: disable=line-too-long
    Single, reusable definition of tests for the interface. Authors of
    new implementations of the interface merely have to provide the test
    data, as class attributes of a class which inherits
    unittest.TestCase AND this class.

    Required test data:
    self.Provider must be set
    self.provider_name must be set
    self.domain must be set
    self._filter_headers can be defined to provide a list of sensitive http headers
    self._filter_query_parameters can be defined to provide a list of sensitive query parameters
    self._filter_post_data_parameters can be defined to provide a list of sensitive post data parameters
    self.provider_variant can be defined as a prefix for saving cassettes when a provider uses multiple variants

    Extended test suites can be skipped by adding the following snippet to the test_{PROVIDER_NAME}.py file

        @pytest.fixture(autouse=True)
        def _skip_suite(self, request):  # pylint: disable=no-self-use
            if request.node.get_marker('ext_suite_1'):
                pytest.skip('Skipping extended suite')

    """

    def __init__(self):
        self.Provider = BaseProvider  # pylint: disable=invalid-name
        self.domain = None
        self.provider_name = None

    ###########################################################################
    # Provider module shape
    ###########################################################################
    def test_provider_module_shape(self):
        module = import_module(
            'lexicon.providers.{0}'.format(
                self.provider_name))

        assert hasattr(module, 'provider_parser')
        assert hasattr(module, 'Provider')
        if self.provider_name != 'auto':
            assert hasattr(module, 'NAMESERVER_DOMAINS')

        assert callable(module.provider_parser)
        assert callable(module.Provider)
        if self.provider_name != 'auto':
            assert isinstance(module.NAMESERVER_DOMAINS, list)

    ###########################################################################
    # Provider.authenticate()
    ###########################################################################
    @_vcr_integration_test
    def test_provider_authenticate(self):
        provider = self._construct_authenticated_provider()
        assert provider.domain_id is not None

    @_vcr_integration_test
    def test_provider_authenticate_with_unmanaged_domain_should_fail(self):
        config = self._test_config()
        config.add_config_source(DictConfigSource(
            {'domain': 'thisisadomainidonotown.com'}), 0)
        provider = self.Provider(config)
        with pytest.raises(Exception):
            provider.authenticate()

    ###########################################################################
    # Provider.create_record()
    ###########################################################################
    @_vcr_integration_test
    def test_provider_when_calling_create_record_for_A_with_valid_name_and_content(self):  # pylint: disable=invalid-name
        provider = self._construct_authenticated_provider()
        assert provider.create_record('A', 'localhost', '127.0.0.1')

    @_vcr_integration_test
    def test_provider_when_calling_create_record_for_CNAME_with_valid_name_and_content(self):  # pylint: disable=invalid-name
        provider = self._construct_authenticated_provider()
        assert provider.create_record('CNAME', 'docs', 'docs.example.com')

    @_vcr_integration_test
    def test_provider_when_calling_create_record_for_TXT_with_valid_name_and_content(self):  # pylint: disable=invalid-name
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', '_acme-challenge.test', 'challengetoken')

    @_vcr_integration_test
    def test_provider_when_calling_create_record_for_TXT_with_full_name_and_content(self):  # pylint: disable=invalid-name
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.full.{0}".format(self.domain), 'challengetoken')

    @_vcr_integration_test
    def test_provider_when_calling_create_record_for_TXT_with_fqdn_name_and_content(self):  # pylint: disable=invalid-name
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.fqdn.{0}.".format(self.domain), 'challengetoken')

    ###########################################################################
    # Provider.list_records()
    ###########################################################################
    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_no_arguments_should_list_all(self):
        provider = self._construct_authenticated_provider()
        assert isinstance(provider.list_records(), list)

    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_name_filter_should_return_record(self):
        provider = self._construct_authenticated_provider()
        provider.create_record('TXT', 'random.test', 'challengetoken')
        records = provider.list_records('TXT', 'random.test')
        assert len(records) == 1
        assert records[0]['content'] == 'challengetoken'
        assert records[0]['type'] == 'TXT'
        assert records[0]['name'] == 'random.test.{0}'.format(self.domain)

    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_full_name_filter_should_return_record(self):
        provider = self._construct_authenticated_provider()
        provider.create_record('TXT', 'random.fulltest.{0}'.format(
            self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'random.fulltest.{0}'.format(self.domain))
        assert len(records) == 1
        assert records[0]['content'] == 'challengetoken'
        assert records[0]['type'] == 'TXT'
        assert records[0]['name'] == 'random.fulltest.{0}'.format(self.domain)

    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_fqdn_name_filter_should_return_record(self):
        provider = self._construct_authenticated_provider()
        provider.create_record('TXT', 'random.fqdntest.{0}.'.format(
            self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'random.fqdntest.{0}.'.format(self.domain))
        assert len(records) == 1
        assert records[0]['content'] == 'challengetoken'
        assert records[0]['type'] == 'TXT'
        assert records[0]['name'] == 'random.fqdntest.{0}'.format(self.domain)

    @_vcr_integration_test
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "ttl.fqdn.{0}.".format(self.domain), 'ttlshouldbe3600')
        records = provider.list_records(
            'TXT', 'ttl.fqdn.{0}'.format(self.domain))
        assert len(records) == 1
        assert str(records[0]['ttl']) == str(3600)

    @pytest.mark.skip(reason="not sure how to test empty list across multiple providers")
    @_vcr_integration_test
    def test_provider_when_calling_list_records_should_return_empty_list_if_no_records_found(self):
        provider = self._construct_authenticated_provider()
        assert isinstance(provider.list_records(), list)

    @pytest.mark.skip(reason="not sure how to test filtering across multiple providers")
    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_arguments_should_filter_list(self):
        provider = self._construct_authenticated_provider()
        assert isinstance(provider.list_records(), list)

    ###########################################################################
    # Provider.update_record()
    ###########################################################################
    @_vcr_integration_test
    def test_provider_when_calling_update_record_should_modify_record(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record('TXT', 'orig.test', 'challengetoken')
        records = provider.list_records('TXT', 'orig.test')
        assert provider.update_record(records[0].get(
            'id', None), 'TXT', 'updated.test', 'challengetoken')

    @_vcr_integration_test
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'orig.nameonly.test', 'challengetoken')
        assert provider.update_record(
            None, 'TXT', 'orig.nameonly.test', 'updated')

    @_vcr_integration_test
    def test_provider_when_calling_update_record_with_full_name_should_modify_record(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'orig.testfull.{0}'.format(self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'orig.testfull.{0}'.format(self.domain))
        assert provider.update_record(records[0].get(
            'id', None), 'TXT', 'updated.testfull.{0}'.format(self.domain), 'challengetoken')

    @_vcr_integration_test
    def test_provider_when_calling_update_record_with_fqdn_name_should_modify_record(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'orig.testfqdn.{0}.'.format(self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'orig.testfqdn.{0}.'.format(self.domain))
        assert provider.update_record(records[0].get(
            'id', None), 'TXT', 'updated.testfqdn.{0}.'.format(self.domain), 'challengetoken')

    ###########################################################################
    # Provider.delete_record()
    ###########################################################################
    @_vcr_integration_test
    def test_provider_when_calling_delete_record_by_identifier_should_remove_record(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record('TXT', 'delete.testid', 'challengetoken')
        records = provider.list_records('TXT', 'delete.testid')
        assert provider.delete_record(records[0]['id'])
        records = provider.list_records('TXT', 'delete.testid')
        assert not records

    @_vcr_integration_test
    def test_provider_when_calling_delete_record_by_filter_should_remove_record(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'delete.testfilt', 'challengetoken')
        assert provider.delete_record(
            None, 'TXT', 'delete.testfilt', 'challengetoken')
        records = provider.list_records('TXT', 'delete.testfilt')
        assert not records

    @_vcr_integration_test
    def test_provider_when_calling_delete_record_by_filter_with_full_name_should_remove_record(self):  # pylint: disable=line-too-long
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'delete.testfull.{0}'.format(self.domain), 'challengetoken')
        assert provider.delete_record(
            None, 'TXT', 'delete.testfull.{0}'.format(self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'delete.testfull.{0}'.format(self.domain))
        assert not records

    @_vcr_integration_test
    def test_provider_when_calling_delete_record_by_filter_with_fqdn_name_should_remove_record(self):  # pylint: disable=line-too-long
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', 'delete.testfqdn.{0}.'.format(self.domain), 'challengetoken')
        assert provider.delete_record(
            None, 'TXT', 'delete.testfqdn.{0}.'.format(self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', 'delete.testfqdn.{0}.'.format(self.domain))
        assert not records

    ###########################################################################
    # Extended Test Suite 1 - March 2018 - Validation for Create Record NOOP & Record Sets
    ###########################################################################

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_create_record_with_duplicate_records_should_be_noop(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.noop.{0}.".format(self.domain), 'challengetoken')
        assert provider.create_record(
            'TXT', "_acme-challenge.noop.{0}.".format(self.domain), 'challengetoken')
        records = provider.list_records(
            'TXT', "_acme-challenge.noop.{0}.".format(self.domain))
        assert len(records) == 1

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_create_record_multiple_times_should_create_record_set(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.createrecordset.{0}.".format(self.domain), 'challengetoken1')
        assert provider.create_record(
            'TXT', "_acme-challenge.createrecordset.{0}.".format(self.domain), 'challengetoken2')

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_list_records_with_invalid_filter_should_be_empty_list(self):
        provider = self._construct_authenticated_provider()
        records = provider.list_records(
            'TXT', 'filter.thisdoesnotexist.{0}'.format(self.domain))
        assert not records

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_list_records_should_handle_record_sets(self):
        provider = self._construct_authenticated_provider()
        provider.create_record(
            'TXT', "_acme-challenge.listrecordset.{0}.".format(self.domain), 'challengetoken1')
        provider.create_record(
            'TXT', "_acme-challenge.listrecordset.{0}.".format(self.domain), 'challengetoken2')
        records = provider.list_records(
            'TXT', '_acme-challenge.listrecordset.{0}.'.format(self.domain))
        assert len(records) == 2

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_delete_record_with_record_set_name_remove_all(self):
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.deleterecordset.{0}.".format(self.domain), 'challengetoken1')
        assert provider.create_record(
            'TXT', "_acme-challenge.deleterecordset.{0}.".format(self.domain), 'challengetoken2')

        assert provider.delete_record(
            None, 'TXT', '_acme-challenge.deleterecordset.{0}.'.format(self.domain))
        records = provider.list_records(
            'TXT', '_acme-challenge.deleterecordset.{0}.'.format(self.domain))
        assert not records

    @pytest.mark.ext_suite_1
    @_vcr_integration_test
    def test_provider_when_calling_delete_record_with_record_set_by_content_should_leave_others_untouched(self):  # pylint: disable=line-too-long
        provider = self._construct_authenticated_provider()
        assert provider.create_record(
            'TXT', "_acme-challenge.deleterecordinset.{0}.".format(self.domain), 'challengetoken1')
        assert provider.create_record(
            'TXT', "_acme-challenge.deleterecordinset.{0}.".format(self.domain), 'challengetoken2')

        assert provider.delete_record(
            None, 'TXT', '_acme-challenge.deleterecordinset.{0}.'
            .format(self.domain), 'challengetoken1')
        records = provider.list_records(
            'TXT', '_acme-challenge.deleterecordinset.{0}.'.format(self.domain))
        assert len(records) == 1

    # Private helpers, mimicing the auth_* options provided by the Client
    # http://stackoverflow.com/questions/6229073/how-to-make-a-python-dictionary-that-returns-key-for-keys-missing-from-the-dicti

    def _test_config(self):
        """
        This method construct a ConfigResolver suitable for tests.
        This will resolve any parameters required by Lexicon or the provider in the following order:
            * parameters that matches the ones provided by _test_parameters_overrides
            * parameters that matches existing environment variables at the time of test execution
            * parameters processed throught the lambda provided by _test_fallback_fn.

        See lexicon/providers/base.py for a full list of parameters available.
        You should not override this method. Just override `self.domain`,
        or use _test_parameters_overrides() to configure specific parameters for the tests.

        Any parameters that you expect to be passed to the provider via the cli,
        like --auth_username and --auth_token, will be present during the tests,
        with a 'placeholder_' prefix.

        options['auth_password'] == 'placeholder_auth_password'
        options['auth_username'] == 'placeholder_auth_username'
        options['unique_provider_option'] == 'placeholder_unique_provider_option'

        You can change this behavior by overriding _test_fallback_fn().

        """
        config = ConfigResolver()
        # First we load the overrides
        overrides = self._test_parameters_overrides()
        overrides['domain'] = self.domain
        overrides['provider_name'] = self.provider_name
        config.with_config_source(EngineOverrideConfigSource(overrides))

        # Then we get environment variables
        config.with_env()

        # And finally we provide the fallback function
        config.with_config_source(
            FallbackConfigSource(self._test_fallback_fn()))

        return config

    def _test_parameters_overrides(self):  # pylint: disable=no-self-use
        """
        This method gives an object whose keys are some provider
        or lexicon parameters expected during a test.
        If a parameter match on of the key during a test,
        the associated value will be used authoritatively.

        Example:
        {'auth_token': 'AUTH_TOKEN'} => if the provider require to use auth_token,
                                        its value will be always AUTH_TOKEN.

        By default no value is overriden.
        """
        return {}

    def _test_fallback_fn(self):  # pylint: disable=no-self-use
        """
        This method gives a fallback lambda for any provider parameter that have not been resolved.
        By default it will return 'placeholder_[parameter_name]' for a particular parameter
        (eg. placeholder_auth_token for auth_token).
        """
        return lambda x: 'placeholder_' + x

    def _cassette_path(self, fixture_subpath):
        """
        A path customized for the provider's fixture.
        The default path is, for example:
            {provider}/IntegrationTests
        but if the test is a `provider_variant`, the path is customized to the variant:
            {provider}/{variant_name}-IntegrationTests
        """
        if self.provider_variant:
            return "{0}/{1}-{2}".format(self.provider_name, self.provider_variant, fixture_subpath)
        return "{0}/{1}".format(self.provider_name, fixture_subpath)

    def _construct_authenticated_provider(self):
        """
        Construct a new provider, and authenticate it against the target DNS provider API.
        """
        provider = self.Provider(self._test_config())
        provider.authenticate()
        return provider

    # Optional. Used to identify the test variant, if any.
    provider_variant = None

    def _filter_headers(self):  # pylint: disable=no-self-use
        return []

    def _filter_query_parameters(self):  # pylint: disable=no-self-use
        return []

    def _filter_post_data_parameters(self):  # pylint: disable=no-self-use
        return []

    def _filter_response(self, response):  # pylint: disable=no-self-use
        """Filter any sensitive data out of the providers response. `response`
        is a Python object with the same structure as all the response sections
        in the YAML recordings at tests/fixtures/cassets/[provider]. For the
        sake of sparing you some time the most important values are:

        response['body']['string']: Contains the HTML or JSON response.
        response['headers']: An object whose keys are HTTP header names
                             e.g. response['headers']['content-length'].
        response['status']: An object that contains 'code' and 'message'
                            subkeys representing the HTTP status code and
                            status message.
        """
        return response



"""Integration tests for Henet"""
from unittest import TestCase

from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.henet import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class FooProviderTests(TestCase, IntegrationTests):
    """TestCase for Henet"""
    Provider = Provider
    provider_name = 'henet'
    domain = 'lexicontest.com'

    def _filter_post_data_parameters(self):
        return ['email', 'pass']

    def _filter_headers(self):
        return ['Authorization', 'Cookie']

    def _filter_query_parameters(self):
        return ['pass']



"""Integration tests for SakuraCloud"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.sakuracloud import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from define_tests.TheTests
class SakruaCloudProviderTests(TestCase, IntegrationTests):
    """TestCase for SakuraCloud"""
    Provider = Provider
    provider_name = 'sakuracloud'
    domain = 'example.com'

    def _filter_headers(self):
        return ['Authorization']

    # TODO: this should be enabled
    @pytest.mark.skip(reason="record id is not exists")
    def test_provider_when_calling_delete_record_by_identifier_should_remove_record(self):
        return

    # TODO: the following skipped suite and fixtures should be enabled
    @pytest.fixture(autouse=True)
    def _skip_suite(self, request):  # pylint: disable=no-self-use
        if request.node.get_marker('ext_suite_1'):
            pytest.skip('Skipping extended suite')



"""Integration tests for Linode V4"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.linode4 import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class Linode4ProviderTests(TestCase, IntegrationTests):
    """TestCase for Linode V4"""
    Provider = Provider
    provider_name = 'linode4'
    domain = 'lexicon-test.com'

    def _filter_post_data_parameters(self):
        return []

    def _filter_headers(self):
        return ['Authorization']

    def _filter_query_parameters(self):
        return []

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""Integration tests for Zeit"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.zeit import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class ZeitProviderTests(TestCase, IntegrationTests):
    """TestCase for Zeit"""
    Provider = Provider
    provider_name = 'zeit'
    domain = 'fullm3tal.tech'

    def _filter_headers(self):
        return ['Authorization']

    @pytest.mark.skip(reason="Records TTL are not supported by Zeit DNS")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



""" Ensure that stdout corresponds to the given reference output """
# pylint: disable=missing-docstring
from __future__ import absolute_import
import json

from lexicon import cli


DATA = [
    {'id': 'fake-id', 'type': 'TXT', 'name': 'fake.example.com',
     'content': 'fake', 'ttl': 3600},
    {'id': 'fake2-id', 'type': 'TXT', 'name': 'fake2.example.com',
     'content': 'fake2', 'ttl': 3600}]


def assert_correct_output(capsys, expected_output_lines):
    out, _ = capsys.readouterr()
    assert out.splitlines() == expected_output_lines


def test_output_function_outputs_json_as_table(capsys):
    expected_output_lines = [
        'ID       TYPE NAME              CONTENT TTL ',
        '-------- ---- ----------------- ------- ----',
        'fake-id  TXT  fake.example.com  fake    3600',
        'fake2-id TXT  fake2.example.com fake2   3600',
    ]

    cli.handle_output(DATA, 'TABLE', 'list')
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_json_as_table_with_no_header(capsys):
    expected_output_lines = [
        'fake-id  TXT fake.example.com  fake  3600',
        'fake2-id TXT fake2.example.com fake2 3600',
    ]

    cli.handle_output(DATA, 'TABLE-NO-HEADER', 'list')
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_json_as_json_string(capsys):
    cli.handle_output(DATA, 'JSON', 'list')

    out, _ = capsys.readouterr()
    json_data = json.loads(out)

    assert json_data == DATA


def test_output_function_output_nothing_when_quiet(capsys):
    expected_output_lines = []

    cli.handle_output(DATA, 'QUIET', 'list')
    assert_correct_output(capsys, expected_output_lines)


def test_output_function_outputs_nothing_with_not_a_json_serializable(capsys):
    expected_output_lines = []

    cli.handle_output(object(), 'TABLE', 'list')
    assert_correct_output(capsys, expected_output_lines)

    cli.handle_output(object(), 'TABLE-NO-HEADER', 'list')
    assert_correct_output(capsys, expected_output_lines)

    cli.handle_output(object(), 'JSON', 'list')
    assert_correct_output(capsys, expected_output_lines)



"""Integratoin tests for Zonomi"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.zonomi import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class ZonomiProviderTests(TestCase, IntegrationTests):
    """TestCase for Zonomi"""
    Provider = Provider
    provider_name = 'zonomi'
    domain = 'pcekper.com.ar'

    def _filter_query_parameters(self):
        return ['api_key']

    # TODO: the following skipped suite and fixtures should be enabled
    @pytest.mark.skip(reason="new test, missing recording")
    def test_provider_when_calling_update_record_should_modify_record_name_specified(self):
        return

    @pytest.fixture(autouse=True)
    def _skip_suite(self, request):  # pylint: disable=no-self-use
        if request.node.get_marker('ext_suite_1'):
            pytest.skip('Skipping extended suite')

    def _test_parameters_overrides(self):
        return {'auth_entrypoint': 'zonomi'}



"""Module provider for luadns"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['luadns.com']


def provider_parser(subparser):
    """Configure provider parser for luadns"""
    subparser.add_argument(
        "--auth-username", help="specify email address for authentication")
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")


class Provider(BaseProvider):
    """Provider class for luadns"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.luadns.com/v1'

    def _authenticate(self):
        payload = self._get('/zones')

        domain_info = next(
            (domain for domain in payload if domain['name'] == self.domain), None)

        if not domain_info:
            raise Exception('No domain found')

        self.domain_id = domain_info['id']

    # Create record. If record already exists with the same content, do nothing'

    def _create_record(self, rtype, name, content):
        # check if record already exists
        existing_records = self._list_records(rtype, name, content)
        if len(existing_records) == 1:
            return True

        self._post('/zones/{0}/records'.format(self.domain_id),
                   {'type': rtype,
                    'name': self._fqdn_name(name),
                    'content': content,
                    'ttl': self._get_lexicon_option('ttl')})

        LOGGER.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get('/zones/{0}/records'.format(self.domain_id))

        records = []
        for record in payload:
            processed_record = {
                'type': record['type'],
                'name': self._full_name(record['name']),
                'ttl': record['ttl'],
                'content': record['content'],
                'id': record['id']
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record['type'] == rtype]
        if name:
            records = [record for record in records if record['name']
                       == self._full_name(name)]
        if content:
            records = [
                record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):

        data = {
            'ttl': self._get_lexicon_option('ttl')
        }
        if rtype:
            data['type'] = rtype
        if name:
            data['name'] = self._fqdn_name(name)
        if content:
            data['content'] = content

        self._put(
            '/zones/{0}/records/{1}'.format(self.domain_id, identifier), data)

        LOGGER.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            self._delete(
                '/zones/{0}/records/{1}'.format(self.domain_id, record_id))

        LOGGER.debug('delete_record: %s', True)
        return True

    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    auth=requests.auth.HTTPBasicAuth(self._get_provider_option(
                                        'auth_username'), self._get_provider_option('auth_token')),
                                    headers={
                                        'Content-Type': 'application/json',
                                        'Accept': 'application/json'
                                    })
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path, listdir

version = 'unknown'
with open(path.join(path.dirname(path.abspath(__file__)), 'VERSION'), encoding='utf-8') as version_file:
    version = version_file.read().strip()

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get a list of all the providers
current_filepath = path.join(here, 'lexicon', 'providers')
providers = [path.splitext(f)[0] for f in listdir(current_filepath) if path.isfile(path.join(current_filepath, f))]
providers = list(sorted(set(providers)))
providers.remove('base')
providers.remove('__init__')

# Define optional dependencies for specific providers.
# Each key of the dict should match a provider name.
extras_require = {
    'namecheap': ['PyNamecheap'],
    'route53': ['boto3'],
    'softlayer': ['SoftLayer'],
    'subreg': ['zeep'],
    'transip': ['transip>=0.3.0'],
    'plesk': ['xmltodict'],
    'henet': ['beautifulsoup4'],
    'hetzner': ['dnspython>=1.15.0','beautifulsoup4'],
    'easyname': ['beautifulsoup4'],
    'localzone': ['localzone'],
}

# Add a 'full' extra, gathering all external dependencies for providers
extras_require['full'] = set([dep for deps in extras_require.values() for dep in deps])

# Define dev/test dependencies
extras_require['dev'] = [
    'pytest==3.8.0',
    'pytest-cov==2.6.0',
    'pytest-xdist==1.23.0',
    'python-coveralls==2.9.1',
    'vcrpy==1.13.0',
    'mock==2.0.0',
]

setup(
    name='dns-lexicon',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='Manipulate DNS records on various DNS providers in a standardized/agnostic way',
    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://github.com/AnalogJ/lexicon',

    # Author details
    author='Jason Kulatunga',
    author_email='jason@thesparktree.com',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: Name Service (DNS)',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',

        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='dns lexicon dns-lexicon dehydrated letsencrypt ' + ' '.join(providers),

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'requests[security]',
        'tldextract',
        'future',
        'cryptography',
        'pyyaml',
    ],

    extras_require=extras_require,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'lexicon=lexicon.cli:main',
        ],
    },
    test_suite='tests'
)



"""Integration tests for Plesk"""
from unittest import TestCase

import pytest
from lexicon.tests.providers.integration_tests import IntegrationTests
from lexicon.providers.plesk import Provider


# Hook into testing framework by inheriting unittest.TestCase and reuse
# the tests which *each and every* implementation of the interface must
# pass, by inheritance from integration_tests.IntegrationTests
class PleskProviderTests(TestCase, IntegrationTests):
    """TestCase for Plesk"""
    Provider = Provider
    provider_name = 'plesk'
    domain = 'lexicon-test.com'

    def _filter_headers(self):
        return ['HTTP_AUTH_LOGIN', 'HTTP_AUTH_PASSWD']

    def _test_parameters_overrides(self):
        return {'plesk_server': 'https://quasispace.de:8443'}

    @pytest.mark.skip(reason="can not set ttl when creating/updating records")
    def test_provider_when_calling_list_records_after_setting_ttl(self):
        return



"""
Lexicon PowerDNS Provider

Author: Will Hughes, 2017

API Docs: https://doc.powerdns.com/md/httpapi/api_spec/

Implementation notes:
* The PowerDNS API does not assign a unique identifier to each record in the way
that Lexicon expects. We work around this by creating an ID based on the record
name, type and content, which when taken together are always unique
* The PowerDNS API has no notion of 'create a single record' or 'delete a single
record'. All operations are either 'replace the RRSet with this new set of records'
or 'delete all records for this name and type. Similarly, there is no notion of
'change the content of this record', because records are identified by their name,
type and content.
* The API is very picky about the format of values used when creating records:
** CNAMEs must be fully qualified
** TXT, LOC records must be quoted
This is why the _clean_content and _unclean_content methods exist, to convert
back and forth between the format PowerDNS expects, and the format Lexicon uses
"""
from __future__ import absolute_import
import json
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = []


def provider_parser(subparser):
    """Configure provider parser for powerdns"""
    subparser.add_argument(
        "--auth-token", help="specify token for authentication")
    subparser.add_argument("--pdns-server", help="URI for PowerDNS server")
    subparser.add_argument(
        "--pdns-server-id", help="Server ID to interact with")


class Provider(BaseProvider):
    """Provider class for PowerDNS"""
    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = self._get_provider_option('pdns_server')

        if self.api_endpoint.endswith('/'):
            self.api_endpoint = self.api_endpoint[:-1]

        if not self.api_endpoint.endswith("/api/v1"):
            self.api_endpoint += "/api/v1"

        self.server_id = self._get_provider_option('pdns_server_id')
        if self.server_id is None:
            self.server_id = 'localhost'

        self.api_endpoint += "/servers/" + self.server_id

        self.api_key = self._get_provider_option('auth_token')
        assert self.api_key is not None

        self._zone_data = None

    def zone_data(self):
        """Get zone data"""
        if self._zone_data is None:
            self._zone_data = self._get('/zones/' + self.domain).json()
        return self._zone_data

    def _authenticate(self):
        self.zone_data()
        self.domain_id = self.domain

    def _make_identifier(self, rtype, name, content):  # pylint: disable=no-self-use
        return "{}/{}={}".format(rtype, name, content)

    def _parse_identifier(self, identifier):  # pylint: disable=no-self-use
        parts = identifier.split('/')
        rtype = parts[0]
        parts = parts[1].split('=')
        name = parts[0]
        content = "=".join(parts[1:])
        return rtype, name, content

    def _list_records(self, rtype=None, name=None, content=None):
        records = []
        for rrset in self.zone_data()['rrsets']:
            if (name is None or self._fqdn_name(rrset['name']) == self._fqdn_name(
                    name)) and (rtype is None or rrset['type'] == rtype):
                for record in rrset['records']:
                    if content is None or record['content'] == self._clean_content(rtype, content):
                        records.append({
                            'type': rrset['type'],
                            'name': self._full_name(rrset['name']),
                            'ttl': rrset['ttl'],
                            'content': self._unclean_content(rrset['type'], record['content']),
                            'id': self._make_identifier(rrset['type'],
                                                        rrset['name'], record['content'])
                        })
        LOGGER.debug('list_records: %s', records)
        return records

    def _clean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        elif rtype == "CNAME":
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            content = content.strip('"')
        elif rtype == "CNAME":
            content = self._full_name(content)
        return content

    def _create_record(self, rtype, name, content):
        content = self._clean_content(rtype, content)
        for rrset in self.zone_data()['rrsets']:
            if rrset['name'] == name and rrset['type'] == rtype:
                update_data = rrset
                if 'comments' in update_data:
                    del update_data['comments']
                update_data['changetype'] = 'REPLACE'
                break
        else:
            update_data = {
                'name': name,
                'type': rtype,
                'records': [],
                'ttl': self._get_lexicon_option('ttl') or 600,
                'changetype': 'REPLACE'
            }

        for record in update_data['records']:
            if record['content'] == content:
                return True

        update_data['records'].append({
            'content': content,
            'disabled': False
        })

        update_data['name'] = self._fqdn_name(update_data['name'])

        request = {'rrsets': [update_data]}
        LOGGER.debug('request: %s', request)

        self._patch('/zones/' + self.domain, data=request)
        self._zone_data = None
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is not None:
            rtype, name, content = self._parse_identifier(identifier)

        LOGGER.debug("delete %s %s %s", rtype, name, content)
        if rtype is None or name is None:
            raise Exception("Must specify at least both rtype and name")

        for rrset in self.zone_data()['rrsets']:
            if rrset['type'] == rtype and self._fqdn_name(rrset['name']) == self._fqdn_name(name):
                update_data = rrset
                if 'comments' in update_data:
                    del update_data['comments']
                update_data['changetype'] = 'REPLACE'
                break
        else:
            return True

        new_records = []
        for record in update_data['records']:
            if content is None or self._unclean_content(
                    rtype, record['content']) != self._unclean_content(rtype, content):
                new_records.append(record)

        update_data['name'] = self._fqdn_name(update_data['name'])
        update_data['records'] = new_records

        request = {'rrsets': [update_data]}
        LOGGER.debug('request: %s', request)

        self._patch('/zones/' + self.domain, data=request)
        self._zone_data = None
        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        self._delete_record(identifier)
        return self._create_record(rtype, name, content)

    def _patch(self, url='/', data=None, query_params=None):
        return self._request('PATCH', url, data=data, query_params=query_params)

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(action, self.api_endpoint + url, params=query_params,
                                    data=json.dumps(data),
                                    headers={
                                        'X-API-Key': self.api_key,
                                        'Content-Type': 'application/json'
                                    })
        LOGGER.debug('response: %s', response.text)
        response.raise_for_status()
        return response



"""Provider module for CloudNS"""
from __future__ import absolute_import
import logging

import requests
from lexicon.providers.base import Provider as BaseProvider


LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['cloudns.net']


def provider_parser(subparser):
    """Configure provider parser for CloudNS"""
    identity_group = subparser.add_mutually_exclusive_group()
    identity_group.add_argument(
        "--auth-id", help="specify user id for authentication")
    identity_group.add_argument(
        "--auth-subid", help="specify subuser id for authentication")
    identity_group.add_argument(
        "--auth-subuser", help="specify subuser name for authentication")
    subparser.add_argument(
        "--auth-password", help="specify password for authentication")
    subparser.add_argument("--weight", help="specify the SRV record weight")
    subparser.add_argument("--port", help="specify the SRV record port")


class Provider(BaseProvider):
    """Provider class for CloudNS"""
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.cloudns.net'

    def _authenticate(self):
        payload = self._get('/dns/get-zone-info.json',
                            {'domain-name': self.domain})
        self.domain_id = payload['name']
        LOGGER.debug('authenticate: %s', payload)

    def _create_record(self, rtype, name, content):
        # Skip execution if such a record already exists
        existing_records = self._list_records(rtype, name, content)
        if existing_records:
            return True

        # Build parameters for adding a new record
        params = {
            'domain-name': self.domain_id,
            'record-rtype': rtype,
            'host': self._relative_name(name),
            'record': content
        }
        if self._get_lexicon_option('ttl'):
            params['ttl'] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            params['priority'] = self._get_lexicon_option('priority')
        if self._get_provider_option('weight'):
            params['weight'] = self._get_lexicon_option('weight')
        if self._get_provider_option('port'):
            params['port'] = self._get_lexicon_option('port')

        # Add new record by calling the ClouDNS API
        payload = self._post('/dns/add-record.json', params)
        LOGGER.debug('create_record: %s', payload)

        # Error handling is already covered by self._request
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        # Build parameters to make use of the built-in API filtering
        params = {'domain-name': self.domain_id}
        if rtype:
            params['type'] = rtype
        if name:
            params['host'] = self._relative_name(name)

        # Fetch and parse all records for the given zone
        payload = self._get('/dns/records.json', params)
        payload = payload if not isinstance(payload, list) else {}
        records = []
        for record in payload.values():
            records.append({
                'type': record['type'],
                'name': self._full_name(record['host']),
                'ttl': record['ttl'],
                'content': record['record'],
                'id': record['id']
            })

        # Filter by content manually as API does not support that
        if content:
            records = [
                record for record in records if record['content'] == content]

        # Print records as debug output and return them
        LOGGER.debug('list_records: %s', records)
        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        if not identifier:
            identifier = self._find_record_identifier(rtype, name, None)

        # Build parameters for updating an existing record
        params = {'domain-name': self.domain_id, 'record-id': identifier}
        if name:
            params['host'] = self._relative_name(name)
        if content:
            params['record'] = content
        if self._get_lexicon_option('ttl'):
            params['ttl'] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            params['priority'] = self._get_lexicon_option('priority')
        if self._get_provider_option('weight'):
            params['weight'] = self._get_provider_option('weight')
        if self._get_provider_option('port'):
            params['port'] = self._get_provider_option('port')

        # Update existing record by calling the ClouDNS API
        payload = self._post('/dns/mod-record.json', params)
        LOGGER.debug('update_record: %s', payload)

        # Error handling is already covered by self._request
        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            # Delete existing record by calling the ClouDNS API
            self._post(
                '/dns/delete-record.json', {'domain-name': self.domain_id, 'record-id': record_id})

        LOGGER.debug('delete_record: %s', True)

        # Error handling is already covered by self._request
        return True

    def _build_authentication_data(self):
        if not self._get_provider_option('auth_password'):
            raise Exception(
                'No valid authentication data passed, expected: auth-password')

        if self._get_provider_option('auth_id'):
            return {'auth-id': self._get_provider_option('auth_id'),
                    'auth-password': self._get_provider_option('auth_password')}
        if self._get_provider_option('auth_subid'):
            return {'sub-auth-id': self._get_provider_option(
                'auth_subid'), 'auth-password': self._get_provider_option('auth_password')}
        if self._get_provider_option('auth_subuser'):
            return {'sub-auth-user': self._get_provider_option(
                'auth_subuser'), 'auth-password': self._get_provider_option('auth_password')}
        if (self._get_provider_option('auth_id') or self._get_provider_option('auth_subid')
                or self._get_provider_option('auth_subuser')):
            # All the options were passed with a fallback value, return an empty dictionary.
            return {}
        raise Exception(
            'No valid authentication data passed, expected: auth-id, auth-subid, auth-subuser')

    def _find_record_identifier(self, rtype, name, content):
        records = self._list_records(rtype, name, content)
        LOGGER.debug('records: %s', records)
        if len(records) == 1:
            return records[0]['id']
        raise Exception('Record identifier could not be found.')

    def _request(self, action='GET', url='/', data=None, query_params=None):
        # Set default values for missing arguments
        data = data if data else {}
        query_params = query_params if query_params else {}

        # Merge authentication data into request
        if action == 'GET':
            query_params.update(self._build_authentication_data())
        else:
            data.update(self._build_authentication_data())

        # Fire request against ClouDNS API and parse result as JSON
        response = requests.request(action, self.api_endpoint +
                                    url, params=query_params, data=data)
        response.raise_for_status()
        payload = response.json()

        # Check ClouDNS specific status code and description
        if ('status' in payload and 'statusDescription' in payload
                and payload['status'] != 'Success'):
            raise Exception('ClouDNS API request has failed: ' +
                            payload['statusDescription'])

        # Return payload
        return payload
